#!/usr/bin/env python3
"""

Backup tool for small MediaWiki wikis

Copyright 2014 Damian Yerrick
Copying and distribution of this file, with or without modification,
are permitted in any medium without royalty provided the copyright
notice and this notice are preserved.  This file is offered as-is,
without any warranty.

"""

from contextlib import closing
import sqlite3

# DY ported to python3 for SNI support on 2014-04-11
# DY factored MWSite into a class

nesdev_sitespec = {
    'api': 'http://wiki.nesdev.com/w/api.php',
    'index': 'http://wiki.nesdev.com/w/index.php',
    'db': "nesdevwiki.sqlite",
    'imagedir': 'wiki-images',
    'napTime': 3
}
userAgent = "PinEightBackupBot/14.04.11 (https://pineight.com/contact/)"

# MediaWiki API #####################################################

def thrice(func, *args, **kwargs):
    tries = 3
    while True:
        try:
            return func(*args, **kwargs)
        except Exception:
            tries -= 1
            if tries < 1:
                raise

class MWSite(object):

    def __init__(self, sitespec):
        self.sitespec = sitespec
        self.last_fetch_time = self.namespaces = self.siteinfo = None

    def nap(self):
        import time
        cur_time = time.time()
        if self.last_fetch_time is not None:
            t = max(0, self.last_fetch_time + self.sitespec['napTime'] - cur_time)
            time.sleep(t)
            cur_time += t
        self.last_fetch_time = cur_time

    def get_url(self, endpoint, args):
        from urllib.parse import urlencode as hbq
        import urllib.request, urllib.error, urllib.parse

        url = '?'.join((endpoint, hbq(args))) if args else endpoint
        headers = {'User-Agent': userAgent}
        req = urllib.request.Request(url, None, headers)
        self.nap()
        with closing(thrice(urllib.request.urlopen, req)) as infp:
            encoding = infp.headers.get_content_charset()
            body = infp.read()
        return body, encoding

    def get_html(self, title):
        q = {'title': title.encode('utf-8'), 'action': 'render'}
        return self.get_url(self.sitespec['index'], q)

    def do_query(self, args):
        import json
        from collections import Mapping

        if isinstance(args, Mapping):
            args = iter(args.items())
        q = [("action", "query"), ("format", "json")]
        q.extend(args)
        body, encoding = self.get_url(self.sitespec['api'], q)
        body = str(body, encoding)
        # JSON in python 3 has become stricter about requiring str
        # as opposed to automatically encoding
        return json.loads(body)

    def get_siteinfo(self):
        if self.siteinfo and self.namespaces:
            return self.siteinfo
        args = {'meta': 'siteinfo',
                'siprop': 'general|namespaces'}
        siteinfo = self.do_query(args)["query"]
        self.siteinfo = siteinfo["general"]
        self.namespaces = siteinfo['namespaces']
        return self.siteinfo

    def get_allpages(self, namespace=0, filters=None):
        """Get a list of (pageid, title) for all pages on the wiki.

namespace -- the id of the namespace to iterate
From https://www.mediawiki.org/wiki/API:Allpages
"You can only enumerate one namespace at a time. By default, the main
namespace will be enumerated"
filters -- dict or list of extra query arguments, such as apfilterredir

"""
        if namespace is not None:
            namespace = int(namespace)
            if namespace < 0:
                return []
        from collections import Mapping
        if isinstance(filters, Mapping):
            filters = iter(filters.items())
        fromids = []
        nextPage = ''
        while nextPage is not None:
            args = {"list": "allpages",
                    "aplimit": 500,
                    "apfrom": nextPage,
                    "apnamespace": namespace}
            if filters:
                args.update(filters)
            
            result = self.do_query(args)
            if "query-continue" in result:
                nextPage = result["query-continue"]["allpages"]["apfrom"]
            else:
                nextPage = None
            allpages = result["query"]["allpages"]
            fromids.extend((row['pageid'], row['title']) for row in allpages)
        return fromids

    def get_allimages(self, aiprop=None):
        ims = []
        nextPage = ''
        if aiprop:
            if isinstance(aiprop, str):
                aiprop = [aiprop]
        while nextPage is not None:
            args = {"list": "allimages",
                    "ailimit": 500,
                    "aifrom": nextPage}
            if aiprop:
                args['aiprop'] = '|'.join(aiprop)
            result = self.do_query(args)
            if "query-continue" in result:
                nextPage = result["query-continue"]["allimages"]["aifrom"]
            else:
                nextPage = None
            ims.extend(result["query"]["allimages"])
        return ims
    

    def get_redirects(self):
        """Get information on all redirects from a MediaWiki site.

Return a tuple (d, fromids, toids):
d -- dictionary from redirect titles to target titles
fromids -- list of (pageid, name) pairs of redirect titles
toids -- list of (pageid, name) pairs of target titles

"""
        fromids = []
        filters = {"apfilterredir": "redirects"}
        self.get_siteinfo()
        for ns in self.namespaces:
            fromids.extend(self.get_allpages(ns, filters=filters))
        targets = {}
        toids = []
        numResults = 40
        batches = (fromids[i:i + numResults]
                   for i in range(0, len(fromids), numResults))
        print("downloading %d redirects" % len(fromids))
        for batch in batches:
            args = {"redirects": "1",
                    "titles": "|".join(row[1] for row in batch)}
            result = self.do_query(args)
            targets.update((row['from'], row['to'])
                           for row in result['query']['redirects'])

            # Broken redirects show up as a "missing" result with no
            # "pageid".  Because toids is just a way of discovering
            # pages more quickly if tasks['newpages'] is false, it's
            # fine to just discard them.
            qpages = result['query']['pages']
            toids.extend((row['pageid'], row['title'])
                         for row in qpages.values()
                         if 'missing' not in row)
        return (targets, fromids, toids)

    def get_page(self, title):
        numResults = 500
        lims = [(arg+'limit', numResults)
                for arg in ['el', 'pl', 'cl', 'tl', 'im']]

        args = {'titles': title.encode('utf-8'),
                'prop': 'categories|extlinks|iwlinks|images|links|revisions|templates',
                'rvprop': 'content'}
        args.update(lims)
        result = self.do_query(args)
        page = next(iter(result['query']['pages'].values()))

        if 'missing' in page and 'pageid' not in page:
            # page must have been deleted or never created in the first place
            return None

        pageid = page['pageid']
        title = page['title']
        markup = next(iter(page['revisions'][0].values()))
        templates = []
        links = []
        iwlinks = []
        els = []
        cats = []
        images = []
    
        while True:
            templates.extend(row['title']
                             for row in page.get('templates', []))
            iwlinks.extend(':'.join((row['prefix'], row['*']))
                           for row in page.get('iwlinks', []))
            links.extend(row['title']
                         for row in page.get('links', []))
            cats.extend(row['title']
                        for row in page.get('categories', []))
            images.extend(row['title']
                          for row in page.get('images', []))
            els.extend(next(iter(row.values()))
                       for row in page.get('extlinks', []))

            # If some were cut off by numResults, try again to get the rest
            qc = result.get('query-continue')
            if not qc:
                break
            args = {'titles': title,
                    'prop': '|'.join(sorted(qc.keys()))}
            args.update(lims)
            for d in qc.values():
                args.update(d)
            result = self.do_query(args)
            page = next(iter(result['query']['pages'].values()))

        # Return all information gathered with API.
        # If you want the markup, use the render endpoint.
        out = {'pageid': pageid, 'title': title, 'templates': templates,
               'images': images, 'categories': cats, 'links': links,
               'iwlinks': iwlinks, 'extlinks': els, 'markup': markup}
        return out

    def get_touched(self, pages):
        """Get the touched dates for a sequence of pages."""

        numResults = 40
        if isinstance(pages, str):
            pages = [pages]
        batches = (pages[i:i + numResults]
                   for i in range(0, len(pages), numResults))
        out = []
        for batch in batches:
            args = {"prop": "info",
                    "titles": "|".join(batch)}
            result = self.do_query(args)
            result = [(row['title'], row.get('touched'))
                      for row in result['query']['pages'].values()]
            
            out.extend((title,
                        touched and touched.rstrip('Z').replace('T', ' '))
                       for (title, touched) in result)
        return out

# Database access ###################################################

kill_seen_unseen_stmt = """
DELETE FROM unseen_pages
WHERE title in (SELECT title FROM pages)
"""
schema = [
"""
CREATE TABLE IF NOT EXISTS a1 (
  steak TEXT
)
""","""
CREATE TABLE IF NOT EXISTS pages (
  pageid INTEGER PRIMARY KEY,
  title TEXT UNIQUE,
  modDate DATE NOT NULL,
  markup TEXT NOT NULL DEFAULT '',
  html TEXT NOT NULL DEFAULT ''
)
""","""
CREATE INDEX IF NOT EXISTS pages_by_date
ON pages (modDate)
""","""
CREATE TABLE IF NOT EXISTS unseen_pages (
  title TEXT PRIMARY KEY
)
""","""
CREATE TABLE IF NOT EXISTS links (
  from_pageid INTEGER NOT NULL,
  manner TEXT NOT NULL,
  to_title TEXT NOT NULL,
  UNIQUE (from_pageid, manner, to_title)
)
""","""
CREATE TABLE IF NOT EXISTS images (
  name TEXT PRIMARY KEY,
  width INTEGER NOT NULL,
  height INTEGER NOT NULL,
  sha1 BINARY NOT NULL
)
""","""
CREATE INDEX IF NOT EXISTS links_by_target
ON links (to_title)
""",
    kill_seen_unseen_stmt
]
# the link manners are wick (internal link), out (external link),
# cat (category membership), inc (template transclusion),
# file (image transclusion), and redirect.

def schedule_first_crawl(db, pagetitle):
    """If a page was never crawled, schedule it to be crawled for the first time."""
    with closing(db.cursor()) as c:
        c.execute('SELECT pageid FROM pages WHERE title = ?',
                  (pagetitle,))
        rows = c.fetchone()
        if not rows:
            s = """
INSERT OR IGNORE INTO unseen_pages (title) VALUES (?)
"""
            c.execute(s, (pagetitle,))

def update_redirects(db, mw):
    import json

    redirdata = mw.get_redirects()
    (targets, fromids, toids) = redirdata
    print("%d redirects point to %d unique pages"
          % (len(fromids), len(toids)))

    # Insert redirect pages
    stmt = """
REPLACE INTO pages (pageid, title, modDate, markup, html)
VALUES (?, ?, CURRENT_TIMESTAMP, ?, '')
"""
    rows = [(pageid, title, "#REDIRECT [[%s]]" % targets[title])
            for (pageid, title) in fromids]
    with closing(db.cursor()) as c:
        c.executemany(stmt, rows)

    # Replace all redirect links
    with closing(db.cursor()) as c:
        c.execute("DELETE FROM links WHERE manner='redirect'")
    stmt = """
INSERT INTO links (from_pageid, manner, to_title)
VALUES (?, 'redirect', ?)
"""
    rows = [(pageid, targets[title])
            for (pageid, title) in fromids]
    with closing(db.cursor()) as c:
        c.executemany(stmt, rows)

    # Prepare to fetch pages redirected to
    for (pageid, title) in toids:
        schedule_first_crawl(db, title)

def queue_new_pages(db, mw):
    filters = {"apfilterredir": "nonredirects"}
    for ns in mw.namespaces:
        pageids = mw.get_allpages(namespace=ns, filters=filters)
        for (pageid, title) in pageids:
            schedule_first_crawl(db, title)

def get_unseen_pages(db):
    with closing(db.cursor()) as c:
        c.execute(kill_seen_unseen_stmt)
        c.execute("SELECT title FROM unseen_pages")
        return [row[0] for row in c.fetchall()]

def get_outdated_pages(db, mw):
    stmt1 = """
SELECT title, modDate FROM pages
WHERE modDate < DATETIME(CURRENT_TIMESTAMP, '-12 hours')
"""
    with closing(db.cursor()) as c:
        c.execute(kill_seen_unseen_stmt)
        c.execute(stmt1)
        saved_moddates = dict(c.fetchall())
    print("%d pages over 12 hours old" % len(saved_moddates))
    mw_moddates = mw.get_touched(list(saved_moddates))
    return [title for (title, touched) in mw_moddates
            if touched > saved_moddates.get(title, '')]

def get_images_to_replace(db, mw):
    mw_images = mw.get_allimages(aiprop='url|sha1|dimensions')
    mw_images = [(row['name'], row['sha1'], row['width'], row['height'], row['url'])
                 for row in mw_images]

    with closing(db.cursor()) as c:
        c.execute("SELECT name, sha1 FROM images")
        saved_sha1s = dict(c.fetchall())
    to_replace =  [(name, sha1, width, height, url)
                   for (name, sha1, width, height, url) in mw_images
                   if saved_sha1s.get(name) != sha1]
    return to_replace

def update_page(db, mw, title):
    mannerCodes = [('templates', 'inc'), ('links', 'wick'), ('iwlinks', 'iw'),
                   ('extlinks', 'out'), ('images', 'file'), ('categories', 'cat')]
    page_stmt = """
REPLACE INTO pages (pageid, title, modDate, markup, html)
VALUES (?, ?, CURRENT_TIMESTAMP, ?, ?)
"""
    link_kill_stmt = """
DELETE FROM links WHERE from_pageid = ?
"""
    link_stmt = """
INSERT INTO links (from_pageid, manner, to_title)
VALUES (?, ?, ?)
"""
    del_unseen_stmt = """
DELETE FROM unseen_pages WHERE title = ?
"""

    page = mw.get_page(title)
    if page is None:
        print("This page does not exist.")
        with closing(db.cursor()) as c:
            c.execute(del_unseen_stmt, (title,))
        return
    
    html, encoding = mw.get_html(title)
    page_row = (page['pageid'], page['title'], page['markup'],
                str(html, encoding))
    links = []
    for (page_key, manner) in mannerCodes:
        links.extend((page['pageid'], manner, dst)
                     for dst in page[page_key])

    # Once I solved the problem of list=allpages skipping non-main
    # namespaces, I removed this part because it only brought in
    # wanted pages (that is, broken links).
##    for (pageid, manner, dst) in links:
##        if manner != 'out':
##            schedule_first_crawl(db, dst)

    with closing(db.cursor()) as c:
        c.execute(page_stmt, page_row)
    with closing(db.cursor()) as c:
        c.execute(link_kill_stmt, (page['pageid'],))
    with closing(db.cursor()) as c:
        c.executemany(link_stmt, links)
    with closing(db.cursor()) as c:
        c.execute(del_unseen_stmt, (page['title'],))

def update_pages(db, mw, pages):
    from random import shuffle
    shuffle(pages)
    error_streak = 0
    for (i, title) in enumerate(pages):
        print("(%d) %s" % (len(pages) - i, title))
        try:
            update_page(db, mw, title)
            db.commit()
        except Exception as e:
            db.rollback()
            error_streak += 4
            if error_streak >= 10:
                raise
            from traceback import print_exc
            print_exc()
        else:
            error_streak = max(0, error_streak - 1)

def update_image(db, mw, name, sha1, width, height, url):
    from os.path import join as pathjoin
    import urllib.request, urllib.error, urllib.parse
    from time import sleep

    headers = {'User-Agent': userAgent}
    req = urllib.request.Request(url, None, headers)
    mw.nap()

    # Download the image
    with closing(urllib.request.urlopen(req)) as infp:
        body = infp.read()
    with open(pathjoin(mw.sitespec['imagedir'], name), 'wb') as outfp:
        outfp.write(body)
    print("  %d bytes" % len(body))
    body = None

    image_stmt = """
REPLACE INTO images (name, sha1, width, height)
VALUES (?, ?, ?, ?)
"""
    with closing(db.cursor()) as c:
        c.execute(image_stmt, (name, sha1, width, height))

def update_images(db, mw, images):
    from random import shuffle
    shuffle(images)
    for (i, (name, sha1, width, height, url)) in enumerate(images):
        print("(%d) %s" % (len(images) - i, name))
        update_image(db, mw, name, sha1, width, height, url)
        db.commit()

def backup(sitespec, tasks=True):
    mw = MWSite(sitespec)
    with closing(sqlite3.connect(sitespec['db'])) as db:
        
        # First make sure the schema is present
        print("Opening %s" % sitespec['db'])
        with closing(db.cursor()) as c:
            for stmt in schema:
                c.execute(stmt)
        db.commit()

        if tasks is True:
            tasks = set(['redirects', 'newpages', 'updatepages', 'updateimages'])

        if 'redirects' in tasks:
            mw.get_siteinfo()
            print("updating redirects from %d namespaces" % len(mw.namespaces))
            update_redirects(db, mw)
            db.commit()
        if 'newpages' in tasks:
            mw.get_siteinfo()
            print("looking for new pages in %d namespaces" % len(mw.namespaces))
            queue_new_pages(db, mw)
            db.commit()
            update_pages(db, mw, get_unseen_pages(db))
        if 'updatepages' in tasks:
            update_pages(db, mw, get_outdated_pages(db, mw))
        if 'updateimages' in tasks:
            outdated = get_images_to_replace(db, mw)
            print("Replacing %d images" % len(outdated))
            update_images(db, mw, outdated)

        print("Closing %s" % sitespec['db'])
        db.execute("VACUUM")
    # TO DO: Make a gzip file

if __name__=='__main__':
    backup(nesdev_sitespec)
